INHERITED_TEXT = "Predefined Pset"
PREDEFINED_PROPERTY_WINDOW_NAME = "Predefined Properties"
DESCRIPTION = "description"
OPTIONAL = "optional"
CURRENT_PR0JECT_PHASE = "current_project_phase"
PROJECT_PHASES = "ProjectPhases"
AGGREGATION_PSET = "AggregationPsetName"
AGGREGATION_ATTRIBUTE = "AggregationAttributeName"
PREDEFINED_PSETS = "PredefinedPropertySets"
PREDEFINED_PSET = "PredefinedPropertySet"
PROPERTY_SET = "PropertySet"
PROPERTY_SETS = "PropertySets"
IDENT_ATTRIBUTE = "ident_attribute"
ATTRIBUTE = "Attribute"
ATTRIBUTES = "Attributes"
OBJECT = "Object"
OBJECTS = "Objects"
AGGREGATE = "Aggregate"
AGGREGATIONS = "Aggregations"
NAME = "name"
IDENTIFIER = "identifer"
PARENT = "parent"
NONE = "None"
DATA_TYPE = "data_type"
VALUE_TYPE = "value_type"
IS_IDENTIFIER = "is_identifier"
CHILD_INHERITS_VALUE = "child_inherits_value"
IS_CONCEPT = "is_concept"
SCRIPT = "Script"
SCRIPTS = "Scripts"
PROJECT = "Project"
VERSION = "version"
AGGREGATES_TO = "aggregates_to"
AUTHOR = "author"
X_POS = "x_pos"
Y_POS = "y_pos"
IS_ROOT = "root"
CONNECTION = "connection"
NODE = "Node"
NODES = "Nodes"
IFC_MAPPINGS = "IfcMappings"
IFC_MAPPING = "IfcMapping"
CUSTOM_ATTRIBUTES = "CustomAttributes"
CUSTOM_ATTRIBUTE = "CustomAttribute"
ABBREVIATION = "abbreviation"
REVIT_MAPPING = "revit_mapping"
VALUE = "Value"
IGNORE_PSET = "IFC"


import uuid
from xml.etree.ElementTree import Element

from lxml import etree

from ... import classes
from ...constants import value_constants
from .desite import handle_header


def _handle_section(id_dict, aggregation: classes.Aggregation, xml_item: Element) -> None:
    xml_child = etree.SubElement(xml_item, "section")
    id_dict[aggregation] = aggregation.uuid
    xml_child.set("ID", aggregation.uuid)
    xml_child.set("name", aggregation.object.name)
    xml_child.set("pre", "")
    xml_child.set("type", "typeBsGroup")
    xml_child.set("takt", "")

    for child in sorted(aggregation.children, key=lambda x: x.name):
        connection_type = child.parent_connection
        if connection_type == value_constants.AGGREGATION:
            _handle_section(id_dict, child, xml_child)
        else:
            _handle_section(id_dict, child, xml_item)


def _handle_elementsection(xml_parent: Element):
    xml_elementsection = etree.SubElement(xml_parent, "elementSection")
    xml_root = etree.SubElement(xml_elementsection, "section")
    xml_root.set("ID", str(uuid.uuid4()))
    xml_root.set("name", "BS Autogenerated")
    xml_root.set("pre", "")
    xml_root.set("type", "typeBsContainer")
    xml_root.set("takt", "")

    root_objects: list[classes.Aggregation] = [aggreg for aggreg in classes.Aggregation if
                                               aggreg.is_root]

    root_objects.sort(key=lambda x: x.name)

    id_dict = dict()
    for aggreg in root_objects:
        _handle_section(id_dict, aggreg, xml_root)

    return xml_elementsection, id_dict


def _handle_property_type_section(xml_repo) -> dict[str, int]:
    xml_property_type_section = etree.SubElement(xml_repo, "propertyTypeSection")

    attribute_dict = dict()

    i = 1
    for attribute in classes.Attribute:
        # use attribute_text instead of attribute to remove duplicates
        attribute_text = f"{attribute.property_set.name}:{attribute.name}"
        if attribute_text not in attribute_dict:
            xml_ptype = etree.SubElement(xml_property_type_section, "ptype")
            xml_ptype.set("key", str(i))
            xml_ptype.set("name", attribute_text)
            xml_ptype.set("datatype", attribute.data_type)
            xml_ptype.set("unit", "")
            xml_ptype.set("inh", "false")
            attribute_dict[attribute_text] = i
            i += 1

    return attribute_dict


def _handle_property_section(xml_repo: etree.Element, id_dict: dict, attribute_dict: dict) -> None:
    xml_property_section = etree.SubElement(xml_repo, "propertySection")

    for node, ref_id in id_dict.items():
        obj = node.object
        for property_set in obj.property_sets:
            for attribute in property_set.attributes:
                attribute_text = f"{attribute.property_set.name}:{attribute.name}"
                ref_type = attribute_dict[attribute_text]
                xml_property = etree.SubElement(xml_property_section, "property")
                xml_property.set("refID", str(ref_id))
                xml_property.set("refType", str(ref_type))
                if attribute == obj.ident_attrib:
                    xml_property.text = attribute.value[0]
                else:
                    xml_property.text = "füllen!"


def _handle_repository(xml_parent: Element, id_dict: dict[classes.Aggregation, str]) -> None:
    xml_repo = etree.SubElement(xml_parent, "repository")
    xml_id_mapping = etree.SubElement(xml_repo, "IDMapping")

    for i, (item, id_value) in enumerate(id_dict.items()):
        xml_id = etree.SubElement(xml_id_mapping, "ID")
        xml_id.set("k", str(i + 1))
        xml_id.set("v", str(id_value))

    attribute_dict = _handle_property_type_section(xml_repo)
    _handle_property_section(xml_repo, id_dict, attribute_dict)


def _handle_relation_section(xml_parent: Element) -> None:
    xml_relation_section = etree.SubElement(xml_parent, "relationSection")
    etree.SubElement(xml_relation_section, "IDMapping")
    xml_relation = etree.SubElement(xml_relation_section, "relation")
    xml_relation.set("name", "default")


def export_bs(project: classes.Project, path: str) -> None:
    if not path:
        return
    xml_boq_export = handle_header(project.author, "bsExport")
    xml_elementsection, id_dict = _handle_elementsection(xml_boq_export)

    etree.SubElement(xml_boq_export, "linkSection")
    _handle_repository(xml_boq_export, id_dict)
    _handle_relation_section(xml_boq_export)

    tree = etree.ElementTree(xml_boq_export)

    with open(path, "wb") as f:
        tree.write(f, xml_declaration=True, pretty_print=True, encoding="utf-8", method="xml")



from __future__ import annotations

import codecs
import logging
import os
import uuid
from xml.etree.ElementTree import Element

import jinja2
from anytree import AnyNode
from lxml import etree

from .desite import handle_header, output_date_time
from ..bim_collab_zoom.rule import merge_list
from ... import __version__
from ... import classes, constants, Template
from ...constants import json_constants, value_constants

JS_EXPORT = "JS"
TABLE_EXPORT = "TABLE"


def _handle_template(path: str | os.PathLike) -> jinja2.Template:
    file_loader = jinja2.FileSystemLoader(Template.HOME_DIR)
    env = jinja2.Environment(loader=file_loader)
    env.trim_blocks = True
    env.lstrip_blocks = True
    template = env.get_template(path)
    return template


def _add_js_rule(parent: Element, file: codecs.StreamReaderWriter) -> str | None:
    name = os.path.basename(file.name)
    if not name.endswith(".js"):
        return None
    else:
        rule_script = etree.SubElement(parent, "ruleScript")

        name = name.split("_")[1:]
        name = "_".join(name)
        rule_script.set("name", name[:-3])
        rule_script.set("active", "true")
        rule_script.set("resume", "false")

        code = etree.SubElement(rule_script, "code")
        file = file.read()

        code.text = file
        code.text = etree.CDATA(code.text)

    return file


def _handle_element_section(xml_qa_export: Element) -> Element:
    xml_element_section = etree.SubElement(xml_qa_export, "elementSection")
    return xml_element_section


def _handle_container(xml_element_section: Element, text) -> Element:
    container = etree.SubElement(xml_element_section, "container")
    container.set("ID", str(uuid.uuid4()))
    container.set("name", text)
    return container


def _handle_checkrun(xml_container: Element, name: str, author: str = "DesiteRuleCreator") -> Element:
    checkrun = etree.SubElement(xml_container, "checkrun")
    _uuid = str(uuid.uuid4())
    checkrun.set("ID", _uuid)
    checkrun.set("name", name)
    checkrun.set("active", "true")
    checkrun.set("user", str(author))
    checkrun.set("date", str(output_date_time))
    checkrun.set("state", "0")
    checkrun.set("objectsOnly", "1")
    checkrun.set("partsOfComposites", "0")
    checkrun.set("createFailed", "true")
    checkrun.set("createWarnings", "true")
    checkrun.set("createIgnored", "true")
    checkrun.set("createPassed", "true")
    checkrun.set("createUndefined", "false")
    return checkrun


def _init_xml(author: str, name: str, version: str) -> (Element, Element):
    xml_qa_export = handle_header(author, "qaExport")
    xml_element_section = _handle_element_section(xml_qa_export)
    text = f"{name} : {version}"
    xml_container = _handle_container(xml_element_section, text)
    return xml_container, xml_qa_export


def _handle_rule(xml_checkrun: Element, rule_type: str) -> Element:
    rule = etree.SubElement(xml_checkrun, "rule")
    rule.set("type", rule_type)
    if rule_type == "UniquePattern":
        etree.SubElement(rule, "patternList")
        code = etree.SubElement(xml_checkrun, "code")
        code.text = ""

    return rule


def _handle_attribute_rule_list(xml_rule: Element) -> Element:
    attribute_rule_list = etree.SubElement(xml_rule, "attributeRuleList")
    return attribute_rule_list


def _define_xml_elements(author: str, xml_container: Element, name: str) -> (Element, Element):
    xml_checkrun = _handle_checkrun(xml_container, name=name, author=author)
    xml_rule = _handle_rule(xml_checkrun, "Attributes")
    xml_attribute_rule_list = _handle_attribute_rule_list(xml_rule)
    _handle_rule(xml_checkrun, "UniquePattern")

    return xml_checkrun, xml_attribute_rule_list


def _handle_js_rules(xml_attribute_rule_list: Element, starts_with: str) -> None:
    folder = os.path.join(Template.HOME_DIR, constants.FILEPATH_JS)

    for fn in os.listdir(folder):
        if str(fn).startswith(starts_with):
            file = codecs.open(f"{folder}/{fn}", encoding="utf-8")
            _add_js_rule(xml_attribute_rule_list, file)


def _handle_rule_script(xml_attribute_rule_list: Element, name: str) -> Element:
    rule_script = etree.SubElement(xml_attribute_rule_list, "ruleScript")
    rule_script.set("name", name)
    rule_script.set("active", "true")
    rule_script.set("resume", "true")
    return rule_script


def _handle_code(xml_rule_script: Element) -> Element:
    code = etree.SubElement(xml_rule_script, "code")
    return code


def _handle_attribute_rule_tree(xml_rule: Element) -> Element:
    attribute_rule_tree = etree.SubElement(xml_rule, "attributeRuleTree")
    return attribute_rule_tree


def _handle_tree_structure(author: str, required_data_dict: dict, parent_xml_container, parent_node: AnyNode, template,
                           xml_object_dict, export_type: str) -> None:
    def check_basics(node):
        obj: classes.Object = node.obj
        if obj.ident_attrib is None:
            return obj, None, True

        pset_dict = required_data_dict.get(obj)
        if pset_dict is None:
            return obj, None, True
        return obj, pset_dict, False

    def create_container(xml_container, node: AnyNode):
        new_xml_container = _handle_container(xml_container, node.obj.name)
        if export_type == "JS":
            create_js_object(new_xml_container, parent_node)
        elif export_type == "CSV":
            create_table_object(new_xml_container, parent_node)
        for child_node in sorted(node.children, key=lambda x: x.id):
            _handle_tree_structure(author, required_data_dict, new_xml_container, child_node, template, xml_object_dict,
                                   export_type)

    def create_js_object(xml_container, node: AnyNode):
        obj, pset_dict, abort = check_basics(node)
        if abort:
            return
        xml_checkrun = _handle_checkrun(xml_container, obj.name, author)
        xml_rule = _handle_rule(xml_checkrun, "Attributes")
        xml_attribute_rule_list = _handle_attribute_rule_list(xml_rule)
        xml_rule_script = _handle_rule_script(xml_attribute_rule_list, name=obj.name)
        xml_code = _handle_code(xml_rule_script)
        cdata_code = template.render(pset_dict=pset_dict, constants=value_constants,
                                     ignore_pset=json_constants.IGNORE_PSET)
        xml_code.text = cdata_code
        _handle_rule(xml_checkrun, "UniquePattern")

        xml_object_dict[xml_checkrun] = obj

    def create_table_object(xml_container, node: AnyNode):
        obj, pset_dict, abort = check_basics(node)
        if abort:
            return
        xml_checkrun = _handle_checkrun(xml_container, obj.name, author)
        xml_rule = _handle_rule(xml_checkrun, "Attributes")
        xml_attribute_rule_tree = _handle_attribute_rule_tree(xml_rule)
        xml_code = _handle_code(xml_container)

        _handle_rule_items_by_pset_dict(pset_dict, xml_attribute_rule_tree)
        xml_code.text = "<![CDATA[]]>"
        _handle_rule(xml_checkrun, "UniquePattern")

        xml_object_dict[xml_checkrun] = obj

    if parent_node.children:
        create_container(parent_xml_container, parent_node)
    else:
        if export_type == JS_EXPORT:
            create_js_object(parent_xml_container, parent_node)
        elif export_type == TABLE_EXPORT:
            create_table_object(parent_xml_container, parent_node)


def _csv_value_in_list(attribute: classes.Attribute):
    return " ".join(str(val) for val in attribute.value)


def _csv_check_range(attribute: classes.Attribute) -> str:
    sorted_range_list = sorted([[min(v1, v2), max(v1, v2)] for [v1, v2] in attribute.value])
    sorted_range_list = merge_list(sorted_range_list)

    pattern = "||".join(f">={v_min}&&<={v_max}" for v_min, v_max in sorted_range_list)
    return pattern


def _build_basics_rule_item(xml_parent: etree.Element, attribute: classes.Attribute) -> etree.Element:
    xml_attrib = etree.SubElement(xml_parent, "ruleItem")
    xml_attrib.set("ID", attribute.uuid)
    xml_attrib.set("name", f"{attribute.property_set.name}:{attribute.name}##{attribute.data_type}")
    xml_attrib.set("type", "simple")
    return xml_attrib


def _handle_rule_item_attribute(xml_parent: etree.Element, attribute: classes.Attribute):
    xml_attrib = _build_basics_rule_item(xml_parent, attribute)

    if not attribute.value:
        xml_attrib.set("pattern", "*")
        return
    pattern = "*"
    if attribute.data_type in (value_constants.XS_INT, value_constants.XS_LONG, value_constants.XS_DOUBLE):
        if attribute.value_type == value_constants.LIST:
            pattern = _csv_value_in_list(attribute)
        elif attribute.value_type == value_constants.RANGE:
            pattern = _csv_check_range(attribute)
        else:
            logging.error(f"No Function defined for {attribute.name} ({attribute.value_type}x{attribute.data_type}")
            pattern = "*"

    elif attribute.data_type == value_constants.XS_STRING:
        if attribute.value_type == value_constants.FORMAT:
            pattern = " || ".join(attribute.value)
        elif attribute.value_type == value_constants.LIST:
            pattern = " ".join(attribute.value)

    elif attribute.data_type == value_constants.XS_BOOL:
        pattern = "*"
    else:
        logging.error(f"No Function defined for {attribute.name} ({attribute.value_type}x{attribute.data_type}")

    xml_attrib.set("pattern", pattern)


def _handle_rule_item_pset(xml_parent: etree.Element, property_set: classes.PropertySet,
                           attributes: list[classes.Attribute]):
    xml_pset = etree.SubElement(xml_parent, "ruleItem")
    xml_pset.set("ID", property_set.uuid)
    xml_pset.set("name", property_set.name)
    xml_pset.set("type", "group")
    for attribute in attributes:
        _handle_rule_item_attribute(xml_pset, attribute)


def _handle_rule_items_by_pset_dict(pset_dict: dict[classes.PropertySet, list[classes.Attribute]],
                                    attribute_rule_tree: etree.Element):
    for pset, attribute_list in pset_dict.items():
        _handle_rule_item_pset(attribute_rule_tree, pset, attribute_list)


def _handle_object_rules(author: str, required_data_dict: dict, project_tree: AnyNode, base_xml_container: Element,
                         template: jinja2.Template, export_type: str) -> dict[Element, classes.Object]:
    xml_object_dict: dict[Element, classes.Object] = dict()

    root_nodes = project_tree.children

    for root_node in sorted(root_nodes, key=lambda x: x.id):
        _handle_tree_structure(author, required_data_dict, base_xml_container, root_node, template, xml_object_dict,
                               export_type)
    return xml_object_dict


def _handle_data_section(xml_qa_export: Element, xml_checkrun_first: Element,
                         xml_checkrun_obj: dict[Element, classes.Object | None],
                         xml_checkrun_last: Element) -> None:
    def get_name() -> str:
        """Transorms native IFC Attributes like IfcType into desite Attributes"""

        pset_name = obj.ident_attrib.property_set.name
        if pset_name == "IFC":
            return obj.ident_attrib.name

        else:
            return f"{pset_name}:{obj.ident_attrib.name}"

    xml_data_section = etree.SubElement(xml_qa_export, "dataSection")

    check_run_data = etree.SubElement(xml_data_section, "checkRunData")
    check_run_data.set("refID", str(xml_checkrun_first.attrib.get("ID")))
    etree.SubElement(check_run_data, "checkSet")

    for xml_checkrun, obj in xml_checkrun_obj.items():
        check_run_data = etree.SubElement(xml_data_section, "checkRunData")
        check_run_data.set("refID", str(xml_checkrun.attrib.get("ID")))
        if obj is None:
            etree.SubElement(check_run_data, "checkSet")
            continue
        filter_list = etree.SubElement(check_run_data, "filterList")
        xml_filter = etree.SubElement(filter_list, "filter")

        xml_filter.set("name", get_name())
        xml_filter.set("dt", "xs:string")
        pattern = f'"{obj.ident_value}"'
        xml_filter.set("pattern", pattern)

    check_run_data = etree.SubElement(xml_data_section, "checkRunData")
    check_run_data.set("refID", str(xml_checkrun_last.attrib.get("ID")))
    filter_list = etree.SubElement(check_run_data, "filterList")
    xml_filter = etree.SubElement(filter_list, "filter")
    xml_filter.set("name", "Check_State")
    xml_filter.set("dt", "xs:string")
    xml_filter.set("pattern", '"Ungeprüft"')


def _handle_property_section(xml_qa_export: Element) -> None:
    repository = etree.SubElement(xml_qa_export, "repository")
    property_type_section = etree.SubElement(repository, "propertyTypeSection")
    ptype = etree.SubElement(property_type_section, "ptype")

    ptype.set("key", "1")
    ptype.set("name", "Bestandsdaten:Objekttyp")
    ptype.set("datatype", "xs:string")
    ptype.set("unit", "")
    ptype.set("inh", "true")
    etree.SubElement(repository, "propertySection")


def export(project: classes.Project,
           required_data_dict: dict[classes.Object, dict[classes.PropertySet, list[classes.Attribute]]],
           path: str, main_pset: str, main_attribute: str, project_tree: AnyNode = None,
           export_type: str = "JS") -> None:
    """

    :param project:
    :param required_data_dict:
    :param path:
    :param project_tree:
    :param export_type: either JS or TABLE
    :return:
    """

    if project_tree is None:
        project_tree = project.tree()
    template = _handle_template(Template.TEMPLATE)
    xml_container, xml_qa_export = _init_xml(project.author, project.name, project.version)
    xml_checkrun_first, xml_attribute_rule_list = _define_xml_elements(project.author, xml_container, "initial_tests")
    _handle_js_rules(xml_attribute_rule_list, "start")
    xml_checkrun_obj = _handle_object_rules(project.author, required_data_dict, project_tree, xml_container, template,
                                            export_type)
    xml_checkrun_last, xml_attribute_rule_list = _define_xml_elements(project.author, xml_container, "untested")
    _handle_js_rules(xml_attribute_rule_list, "end")
    _handle_untested(xml_attribute_rule_list, main_pset, main_attribute)
    _handle_data_section(xml_qa_export, xml_checkrun_first, xml_checkrun_obj, xml_checkrun_last)
    _handle_property_section(xml_qa_export)

    tree = etree.ElementTree(xml_qa_export)
    with open(path, "wb") as f:
        tree.write(f, xml_declaration=True, pretty_print=True, encoding="utf-8", method="xml")


def _handle_untested(xml_attribute_rule_list: etree.Element, main_pset: str, main_attribute: str):
    template = _handle_template(Template.UNTESTED)
    cdata_code = template.render(pset_name=main_pset, ident_attrib=main_attribute)
    rule_script = etree.SubElement(xml_attribute_rule_list, "ruleScript")
    name = "untested"
    rule_script.set("name", name)
    rule_script.set("active", "true")
    rule_script.set("resume", "false")
    code = etree.SubElement(rule_script, "code")
    code.text = etree.CDATA(cdata_code)


def csv_export(required_data_dict: dict[classes.Object, dict[classes.PropertySet, list[classes.Attribute]]], path):
    lines = list()
    lines.append(";".join(["#", f"Created by SOMcreator v{__version__}"]))
    lines.append("H;Property Name;;Data Type;Rule;Comment")

    for obj, pset_dict in required_data_dict.items():
        if obj.ident_attrib is None:
            continue
        ident_attrib = f"{obj.ident_attrib.property_set.name}:{obj.ident_attrib.name}"
        lines.append(";".join(
            ["C", ident_attrib, "", obj.ident_attrib.data_type, obj.ident_value, f"Nach Objekt {obj.name} filtern"]))

        for pset, attribute_list in pset_dict.items():
            for attribute in attribute_list:
                if attribute.value_type != value_constants.RANGE:
                    lines.append(_handle_attribute_rule(attribute))
    with open(path, "w") as file:
        for line in lines:
            file.write(line + "\n")


def _handle_attribute_rule(attribute: classes.Attribute) -> str:
    if attribute.value_type == value_constants.RANGE:
        return ";".join(["R", "", f"{attribute.property_set.name}:{attribute.name}", attribute.data_type, "*",
                         f"Pruefung"])

    if not attribute.value:
        return ";".join(["R", "", f"{attribute.property_set.name}:{attribute.name}", attribute.data_type, "*",
                         f"Pruefung"])

    return ";".join(
        ["R", "", f"{attribute.property_set.name}:{attribute.name}", attribute.data_type, " ".join(attribute.value),
         f"Pruefung"])


def _fast_object_check(main_pset: str, main_attrib: str, author: str, required_data_dict: dict,
                       base_xml_container: Element,
                       template: jinja2.Template) -> dict[Element, None]:
    xml_object_dict: dict[Element, classes.Object] = dict()
    xml_checkrun = _handle_checkrun(base_xml_container, "Main Check", author)
    xml_rule = _handle_rule(xml_checkrun, "Attributes")
    xml_attribute_rule_list = _handle_attribute_rule_list(xml_rule)
    xml_rule_script = _handle_rule_script(xml_attribute_rule_list, name="Main Check")
    xml_code = _handle_code(xml_rule_script)
    cdata_code = template.render(object_dict=required_data_dict, main_pset=main_pset, main_attrib=main_attrib,
                                 constants=value_constants,
                                 ignore_pset=json_constants.IGNORE_PSET)
    xml_code.text = cdata_code
    _handle_rule(xml_checkrun, "UniquePattern")
    return {xml_checkrun: None}


def fast_check(project: classes.Project, main_pset: str, main_attrib: str,
               required_data_dict: dict[classes.Object, dict[classes.PropertySet, list[classes.Attribute]]],
               path: str) -> None:
    """
    creates a single rule for all elements -> no containers for checkruns
    :param project:
    :param required_data_dict: Dictionary of all required Objects, Propertysets and Attributes
    :param path: Export Path
    :return:
    """
    template = _handle_template(Template.FAST_TEMPLATE)
    xml_container, xml_qa_export = _init_xml(project.author, project.name, project.version)
    xml_checkrun_first, xml_attribute_rule_list = _define_xml_elements(project.author, xml_container, "initial_tests")
    _handle_js_rules(xml_attribute_rule_list, "start")
    xml_checkrun_obj = _fast_object_check(main_pset, main_attrib, project.author, required_data_dict, xml_container,
                                          template)
    xml_checkrun_last, xml_attribute_rule_list = _define_xml_elements(project.author, xml_container, "untested")
    _handle_js_rules(xml_attribute_rule_list, "end")
    _handle_data_section(xml_qa_export, xml_checkrun_first, xml_checkrun_obj, xml_checkrun_last)
    _handle_property_section(xml_qa_export)

    tree = etree.ElementTree(xml_qa_export)
    with open(path, "wb") as f:
        tree.write(f, xml_declaration=True, pretty_print=True, encoding="utf-8", method="xml")



DEFAULT_NS = "{http://standards.buildingsmart.org/IDS}"
SCHEME_LOCATION_NS = "http://standards.buildingsmart.org/IDS/ids_05.xsd"

# Classes
APPLICABILITY = DEFAULT_NS + 'applicability'
APPLICABILITYTYPE = DEFAULT_NS + 'applicabilityType'
ATTRIBUTE = DEFAULT_NS + 'attribute'
ATTRIBUTETYPE = DEFAULT_NS + 'attributeType'
AUTHOR = DEFAULT_NS + 'author'
CLASSIFICATION = DEFAULT_NS + 'classification'
CLASSIFICATIONTYPE = DEFAULT_NS + 'classificationType'
COPYRIGHT = DEFAULT_NS + 'copyright'
DATE = DEFAULT_NS + 'date'
DESCRIPTION = DEFAULT_NS + 'description'
ENTITY = DEFAULT_NS + 'entity'
ENTITYTYPE = DEFAULT_NS + 'entityType'
IDS = DEFAULT_NS + 'ids'
IDSVALUE = DEFAULT_NS + 'idsValue'
INFO = DEFAULT_NS + 'info'
MATERIAL = DEFAULT_NS + 'material'
MATERIALTYPE = DEFAULT_NS + 'materialType'
MILESTONE = DEFAULT_NS + 'milestone'
NAME = DEFAULT_NS + 'name'
PARTOF = DEFAULT_NS + 'partOf'
PARTOFTYPE = DEFAULT_NS + 'partOfType'
PREDEFINEDTYPE = DEFAULT_NS + 'predefinedType'
PROPERTY = DEFAULT_NS + 'property'
PROPERTYSET = DEFAULT_NS + 'propertySet'
PROPERTYTYPE = DEFAULT_NS + 'propertyType'
PURPOSE = DEFAULT_NS + 'purpose'
REQUIREMENTS = DEFAULT_NS + 'requirements'
REQUIREMENTSTYPE = DEFAULT_NS + 'requirementsType'
SIMPLEVALUE = DEFAULT_NS + 'simpleValue'
SPECIFICATION = DEFAULT_NS + 'specification'
SPECIFICATIONTYPE = DEFAULT_NS + 'specificationType'
SPECIFICATIONS = DEFAULT_NS + 'specifications'
SPECIFICATIONSTYPE = DEFAULT_NS + 'specificationsType'
SYSTEM = DEFAULT_NS + 'system'
TITLE = DEFAULT_NS + 'title'
VALUE = DEFAULT_NS + 'value'
VERSION = DEFAULT_NS + 'version'

# Attributes
ATTR_DATATYPE = 'datatype'
ATTR_DESCRIPTION = 'description'
ATTR_IDENTIFIER = 'identifier'
ATTR_IFCVERSION = 'ifcVersion'
ATTR_INSTRUCTIONS = 'instructions'
ATTR_NAME = 'name'
ATTR_RELATION = 'relation'
ATTR_URI = 'uri'

# Values
VAL_IFC2X3 = 'IFC2X3'
VAL_IFC4 = 'IFC4'
VAL_IFC4X3 = 'IFC4X3'
VAL_IFCRELAGGREGATES = 'IFCRELAGGREGATES'
VAL_IFCRELASSIGNSTOGROUP = 'IFCRELASSIGNSTOGROUP'
VAL_IFCRELCONTAINEDINSPATIALSTRUCTURE = 'IFCRELCONTAINEDINSPATIALSTRUCTURE'
VAL_IFCRELFILLSELEMENT = 'IFCRELFILLSELEMENT'
VAL_IFCRELNESTS = 'IFCRELNESTS'
VAL_IFCRELVOIDSELEMENT = 'IFCRELVOIDSELEMENT'
